package ocaml-ai-sdk

  1. Overview
  2. Docs
Legend:
Page
Library
Module
Module type
Parameter
Class
Class type
Module Ai_provider_anthropic.Anthropic_optionsSource

Anthropic-specific provider options.

Sourcetype structured_output_mode =
  1. | Auto
  2. | Output_format
  3. | Json_tool
Sourcetype t = {
  1. thinking : Thinking.t option;
  2. effort : Effort.t option;
  3. tool_streaming : bool;
  4. structured_output_mode : structured_output_mode;
}
Sourceval default : t

Default options: no explicit thinking or effort, tool streaming enabled, auto structured output.

Cache control is configured per content block via Cache_control_options.with_cache_control, not via a top-level option.

Sourcetype Ai_provider.Provider_options.key +=
  1. | Anthropic : t Ai_provider.Provider_options.key
    (*

    GADT key for Anthropic-specific options in Provider_options.t.

    *)
Sourceval to_provider_options : t -> Ai_provider.Provider_options.t

Wrap into a Provider_options.t.

Sourceval of_provider_options : Ai_provider.Provider_options.t -> t option

Extract from a Provider_options.t. Returns None if absent.